# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import errno
import json
import os
import random
import stat
import warnings
from hashlib import sha1

import requests
import six
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from descarteslabs.client.exceptions import AuthError, OauthError

DEFAULT_TOKEN_INFO_PATH = os.path.join(
    os.path.expanduser("~"), '.descarteslabs', 'token_info.json')


def base64url_decode(input):
    """Helper method to base64url_decode a string.
    Args:
        input (str): A base64url_encoded string to decode.
    """
    rem = len(input) % 4
    if rem > 0:
        input += b'=' * (4 - rem)

    return base64.urlsafe_b64decode(input)


def makedirs_if_not_exists(path):
    if not os.path.exists(path):
        try:
            os.makedirs(path)
            os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
        except OSError as ex:
            if ex.errno == errno.EEXIST:
                pass
            else:
                raise


class Auth:
    def __init__(self, domain="https://accounts.descarteslabs.com",
                 scope=None, leeway=500, token_info_path=DEFAULT_TOKEN_INFO_PATH,
                 client_id=None, client_secret=None, jwt_token=None, refresh_token=None):
        """
        Helps retrieve JWT from a client id and refresh token for cli usage.
        :param domain: endpoint for auth0
        :param scope: the JWT fields to be included
        :param leeway: JWT expiration leeway
        :param token_info_path: path to a JSON file optionally holding auth information
        :param client_id: JWT client id
        :param client_secret: JWT client secret
        :param jwt_token: the JWT token, if we already have one
        :param refresh_token: the refresh token
        """
        self.token_info_path = token_info_path

        token_info = {}
        if self.token_info_path:
            try:
                with open(self.token_info_path) as fp:
                    token_info = json.load(fp)
            except (IOError, ValueError):
                pass

        self.client_id = client_id if client_id else os.environ.get('CLIENT_ID', token_info.get('client_id', None))
        self.client_secret = client_secret if client_secret else os.environ.get('CLIENT_SECRET', token_info.get(
            'client_secret', None))
        self.refresh_token = refresh_token if refresh_token \
            else os.environ.get('DESCARTESLABS_REFRESH_TOKEN', token_info.get('refresh_token', None))
        self.scope = scope if scope else token_info.get('scope')
        self._token = jwt_token if jwt_token else os.environ.get('JWT_TOKEN', token_info.get('jwt_token', None))

        if token_info:
            # If the token was read from a path but environment variables were set, we may need
            # to reset the token.
            client_id_changed = token_info.get('client_id', None) != self.client_id
            client_secret_changed = token_info.get('client_secret', None) != self.client_secret
            refresh_token_changed = token_info.get('refresh_token', None) != self.refresh_token

            if client_id_changed or client_secret_changed or refresh_token_changed:
                self._token = None

        self._namespace = None
        self._session = None
        self.domain = domain
        self.leeway = leeway

    @classmethod
    def from_environment_or_token_json(cls, **kwargs):
        """
        Creates an Auth object from environment variables CLIENT_ID, CLIENT_SECRET,
        JWT_TOKEN if they are set, or else from a JSON file at the given path.
        :param domain: endpoint for auth0
        :param scope: the JWT fields to be included
        :param leeway: JWT expiration leeway
        :param token_info_path: path to a JSON file optionally holding auth information
        """
        return Auth(**kwargs)

    @property
    def token(self):
        if self._token is None:
            self._get_token()
        else:  # might have token but could be close to expiration
            exp = self.payload.get('exp')

            if exp is not None:
                now = (datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1)).total_seconds()
                if now + self.leeway > exp:
                    try:
                        self._get_token()
                    except AuthError as e:
                        # Unable to refresh, raise if now > exp
                        if now > exp:
                            raise e

        return self._token

    @property
    def payload(self):
        if self._token is None:
            self._get_token()

        if isinstance(self._token, six.text_type):
            token = self._token.encode('utf-8')
        else:
            token = self._token

        claims = token.split(b'.')[1]
        return json.loads(base64url_decode(claims).decode('utf-8'))

    @property
    def session(self):
        if self._session is None:
            self._session = requests.Session()
            retries = Retry(total=5,
                            backoff_factor=random.uniform(1, 10),
                            method_whitelist=frozenset(['GET', 'POST']),
                            status_forcelist=[429, 500, 502, 503, 504])

            self._session.mount('https://', HTTPAdapter(max_retries=retries))

        return self._session

    def _get_token(self, timeout=100):
        if self.client_id is None:
            raise AuthError("Could not find client_id")

        if self.client_secret is None and self.refresh_token is None:
            raise AuthError("Could not find client_secret or refresh token")

        if self.client_id in ["ZOBAi4UROl5gKZIpxxlwOEfx8KpqXf2c"]:  # TODO(justin) remove legacy handling
            # TODO (justin) insert deprecation warning
            if self.scope is None:
                scope = ['openid', 'name', 'groups', 'org']
            else:
                scope = self.scope
            params = {
                "scope": " ".join(scope),
                "client_id": self.client_id,
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "target": self.client_id,
                "api_type": "app",
                "refresh_token": self.refresh_token if self.refresh_token is not None else self.client_secret
            }
        else:
            params = {
                "client_id": self.client_id,
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token if self.refresh_token is not None else self.client_secret
            }

            if self.scope is not None:
                params["scope"] = " ".join(self.scope)

        r = self.session.post(self.domain + "/token", json=params, timeout=timeout)

        if r.status_code != 200:
            raise OauthError("%s: %s" % (r.status_code, r.text))

        data = r.json()
        access_token = data.get('access_token')
        id_token = data.get('id_token')  # TODO(justin) remove legacy id_token usage

        if access_token is not None:
            self._token = access_token
        elif id_token is not None:
            self._token = id_token
        else:
            raise OauthError("could not retrieve token")
        token_info = {}

        if self.token_info_path:
            try:
                with open(self.token_info_path) as fp:
                    token_info = json.load(fp)
            except (IOError, ValueError):
                pass

        token_info['jwt_token'] = self._token

        if self.token_info_path:
            token_info_directory = os.path.dirname(self.token_info_path)
            makedirs_if_not_exists(token_info_directory)

            try:
                with open(self.token_info_path, 'w+') as fp:
                    json.dump(token_info, fp)

                os.chmod(self.token_info_path, stat.S_IRUSR | stat.S_IWUSR)
            except IOError as e:
                warnings.warn('failed to save token: {}'.format(e))

    @property
    def namespace(self):
        if self._namespace is None:
            self._namespace = sha1(self.payload['sub'].encode('utf-8')).hexdigest()
        return self._namespace


if __name__ == '__main__':
    auth = Auth()

    print(auth.token)



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import mock
import os
from six import StringIO
from descarteslabs.client.scripts.cli import parser, handle
import base64
import json


class TestScripts(unittest.TestCase):
    old_token = None
    token_path = os.path.join(os.path.expanduser("~"), '.descarteslabs', 'token_info.json')

    @classmethod
    def setUpClass(cls):
        client_id = os.environ.get("CLIENT_ID")
        client_secret = os.environ.get("CLIENT_SECRET")
        if os.path.exists(cls.token_path):
            cls.old_token = json.load(open(cls.token_path))
            client_id = client_id or cls.old_token.get('client_id')
            client_secret = client_secret or cls.old_token.get('client_secret')

        cls.token = base64.b64encode(
            json.dumps({
                "client_id": client_id,
                "client_secret": client_secret}).encode(
                'utf-8'))

    @classmethod
    def tearDownClass(cls):
        # Put old token back if it existed.
        if cls.old_token:
            with open(cls.token_path, 'w+') as f:
                json.dump(cls.old_token, f)

    def test_auth_login(self):
        with mock.patch('descarteslabs.client.auth.cli.input', return_value=self.token):
            handle(parser.parse_args(["auth", "login"]))

        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["auth", "groups"]))
            self.assertEqual(out.getvalue().strip(), '["public"]')

    def test_places_find(self):
        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["places", "find", "iowa"]))
            iowa = [
                {
                    "name": "Iowa",
                    "bbox": [
                        - 96.639468,
                        40.37544,
                        -90.140061,
                        43.501128
                    ],
                    "id": 85688713,
                    "path": "continent:north-america_country:united-states_region:iowa",
                    "slug": "north-america_united-states_iowa",
                    "placetype": "region"
                }
            ]
            self.assertEqual(json.loads(out.getvalue().strip()), iowa)

    def test_metadata_sources(self):
        with mock.patch('sys.stdout', new_callable=StringIO) as out:
            handle(parser.parse_args(["metadata", "sources"]))
            sources = [{'sat_id': 'LANDSAT_8', 'product': 'landsat:LC08:PRE:TOAR'}]
            self.assertEqual(json.loads(out.getvalue().strip()), sources)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import six
import unittest

import responses
from mock import patch

from descarteslabs.client.auth import Auth


def token_response_callback(request):
    body = request.body
    if not isinstance(body, six.text_type):
        body = body.decode('utf-8')

    data = json.loads(body)

    required_fields = ['client_id', 'grant_type', 'refresh_token']
    legacy_required_fields = ["api_type", "target"]

    if not all(field in data for field in required_fields):
        return 400, {"Content-Type": "application/json"}, json.dumps("missing fields")

    if data['grant_type'] == "urn:ietf:params:oauth:grant-type:jwt-bearer" \
            and all(field in data for field in legacy_required_fields):
        return 200, {"Content-Type": "application/json"}, json.dumps(dict(id_token="id_token"))

    if data['grant_type'] == "refresh_token" \
            and all(field not in data for field in legacy_required_fields):
        return 200, {"Content-Type": "application/json"}, json.dumps(dict(access_token="access_token",
                                                                          id_token="id_token"))
    return 400, {"Content-Type": "application/json"}, json.dumps(data)


class TestAuth(unittest.TestCase):
    @responses.activate
    def test_get_token(self):
        responses.add(responses.POST, 'https://accounts.descarteslabs.com/token',
                      json=dict(access_token="access_token"),
                      status=200)
        auth = Auth(token_info_path=None, client_secret="client_secret", client_id="client_id")
        auth._get_token()

        self.assertEqual("access_token", auth._token)

    @responses.activate
    def test_get_token_legacy(self):
        responses.add(responses.POST, 'https://accounts.descarteslabs.com/token',
                      json=dict(id_token="id_token"), status=200)
        auth = Auth(token_info_path=None, client_secret="client_secret", client_id="client_id")
        auth._get_token()

        self.assertEqual("id_token", auth._token)

    @patch("descarteslabs.client.auth.Auth.payload", new=dict(sub="asdf"))
    def test_get_namespace(self):
        auth = Auth(token_info_path=None, client_secret="client_secret", client_id="client_id")
        self.assertEqual(auth.namespace, "3da541559918a808c2402bba5012f6c60b27661c")

    def test_init_token_no_path(self):
        auth = Auth(jwt_token="token", token_info_path=None, client_id="foo")
        self.assertEqual("token", auth._token)

    @responses.activate
    def test_get_token_schema_internal_only(self):
        responses.add_callback(responses.POST, 'https://accounts.descarteslabs.com/token',
                               callback=token_response_callback)
        auth = Auth(token_info_path=None, refresh_token="refresh_token", client_id="client_id")
        auth._get_token()

        self.assertEqual("access_token", auth._token)

        auth = Auth(token_info_path=None, client_secret="refresh_token", client_id="client_id")
        auth._get_token()

        self.assertEqual("access_token", auth._token)

    @responses.activate
    def test_get_token_schema_legacy_internal_only(self):
        responses.add_callback(responses.POST, 'https://accounts.descarteslabs.com/token',
                               callback=token_response_callback)
        auth = Auth(token_info_path=None, client_secret="client_secret", client_id="ZOBAi4UROl5gKZIpxxlwOEfx8KpqXf2c")
        auth._get_token()
        self.assertEqual("id_token", auth._token)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
import os
from functools import partial
from cachetools import TTLCache, cachedmethod
from cachetools.keys import hashkey
from descarteslabs.client.services.service import Service
from descarteslabs.client.auth import Auth
from six import string_types
from descarteslabs.common.dotdict import DotDict, DotList


class Places(Service):
    TIMEOUT = (9.5, 360)
    """Places and statistics service"""

    def __init__(self, url=None, auth=None, maxsize=10, ttl=600):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if auth is None:
            auth = Auth()

        if url is None:
            url = os.environ.get("DESCARTESLABS_PLACES_URL", "https://platform.descarteslabs.com/waldo/v2")

        super(Places, self).__init__(url, auth=auth)
        self.cache = TTLCache(maxsize, ttl)

    def placetypes(self):
        """Get a list of place types.

        return: list
            List of placetypes ['continent', 'country', 'dependency', 'macroregion', 'region',
                                'district', 'mesoregion', 'microregion', 'county', 'locality']
        """
        r = self.session.get('/placetypes')
        return r.json()

    def random(self, geom='low', placetype=None):
        """Get a random location

        geom: string
            Resolution for the shape [low (default), medium, high]

        return: geojson
        """
        params = {}

        if geom:
            params['geom'] = geom

        if placetype:
            params['placetype'] = placetype

        r = self.session.get('/random', params=params)

        if r.status_code != 200:
            raise RuntimeError("%s: %s" % (r.status_code, r.text))

        return DotDict(r.json())

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'find'))
    def find(self, path, **kwargs):
        """Find candidate slugs based on full or partial path.

        :param str path: Candidate underscore-separated slug.
        :param placetype: Optional place type for filtering.

        Example::

            >>> from descarteslabs.client.services import Places
            >>> results = Places().find('morocco')
            >>> _ = results[0].pop('bbox')
            >>> results
            [
              {
                'id': 85632693,
                'name': 'Morocco',
                'path': 'continent:africa_country:morocco',
                'placetype': 'country',
                'slug': 'africa_morocco'
              }
            ]
        """
        r = self.session.get('/find/%s' % path, params=kwargs)
        return DotList(r.json())

    def search(self, q, limit=10, country=None, region=None, placetype=None, **params):
        """Search for shapes

        :param str q: A query string.
        :param int limit: Max number of matches to return
        :param str country: Restrict search to a specific country
        :param str region: Restrict search to a specific region
        :param str placetype: Restrict search to a specific placetype

        :return: list of candidates

        Example::
            >>> from descarteslabs.client.services import Places
            >>> results = Places().search('texas')
            >>> results[0]
            {
              'bbox': [-106.645584, 25.837395, -93.508039, 36.50035],
              'id': 85688753,
              'name': 'Texas',
              'placetype': 'region',
              'slug': 'north-america_united-states_texas'
            }
        """
        if q:
            params['q'] = q

        if country:
            params['country'] = country

        if region:
            params['region'] = region

        if placetype:
            params['placetype'] = placetype

        if limit:
            params['n'] = limit

        r = self.session.get('/search', params=params, timeout=self.TIMEOUT)

        return DotList(r.json())

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'shape'))
    def shape(self, slug, output='geojson', geom='low', **params):
        """Get the geometry for a specific slug

        :param slug: Slug identifier.
        :param str output: Desired geometry format (`GeoJSON`).
        :param str geom: Desired resolution for the geometry (`low`, `medium`, `high`).

        :return: GeoJSON ``Feature``

        Example::
            >>> from descarteslabs.client.services import Places
            >>> kansas = Places().shape('north-america_united-states_kansas')
            >>> kansas['bbox']
            [-102.051744, 36.993016, -94.588658, 40.003078]

            >>> kansas['geometry']['type']
            'Polygon'

            >>> kansas['properties']
            {
              'name': 'Kansas',
              'parent_id': 85633793,
              'path': 'continent:north-america_country:united-states_region:kansas',
              'placetype': 'region',
              'slug': 'north-america_united-states_kansas'
            }

        """
        params['geom'] = geom
        r = self.session.get('/shape/%s.%s' % (slug, output), params=params)
        return DotDict(r.json())

    @cachedmethod(operator.attrgetter('cache'), key=partial(hashkey, 'prefix'))
    def prefix(self, slug, output='geojson', placetype=None, geom='low', **params):
        """Get all the places that start with a prefix

        :param str slug: Slug identifier.
        :param str output: Desired geometry format (`GeoJSON`, `TopoJSON`).
        :param str placetype: Restrict results to a particular place type.
        :param str geom: Desired resolution for the geometry (`low`, `medium`, `high`).

        :return: GeoJSON or TopoJSON ``FeatureCollection``

        Example::
            >>> from descarteslabs.client.services import Places
            >>> il_counties = Places().prefix('north-america_united-states_illinois', placetype='county')
            >>> len(il_counties['features'])
            102

        """
        if placetype:
            params['placetype'] = placetype
        params['geom'] = geom
        r = self.session.get('/prefix/%s.%s' % (slug, output), params=params)

        return DotDict(r.json())

    def sources(self):
        """Get a list of sources
        """
        r = self.session.get('/sources', timeout=self.TIMEOUT)

        return DotList(r.json())

    def categories(self):
        """Get a list of categories
        """
        r = self.session.get('/categories', timeout=self.TIMEOUT)

        return DotList(r.json())

    def metrics(self):
        """Get a list of metrics
        """
        r = self.session.get('/metrics', timeout=self.TIMEOUT)

        return DotList(r.json())

    def data(self, slug, source=None, category=None, metric=None, units=None, date=None, placetype='county', **params):
        """Get all values for a prefix search and point in time

        :param str slug: Slug identifier (or shape id).
        :param str source: Source
        :param str category: Category
        :param str metric: Metric
        :param str units: Units
        :param str date: Date
        :param str placetype: Restrict results to a particular place type.

        """
        if source:
            params['source'] = source

        if category:
            params['category'] = category

        if metric:
            params['metric'] = metric

        if units:
            params['units'] = units

        if date:
            params['date'] = date

        if placetype:
            params['placetype'] = placetype

        r = self.session.get('/data/%s' % (slug),
                             params=params, timeout=self.TIMEOUT)

        return r.json()

    def statistics(self, slug, source=None, category=None, metric=None, units=None, **params):
        """Get a time series for a specific place

        :param str slug: Slug identifier (or shape id).
        :param str slug: Slug identifier (or shape id).
        :param str source: Source
        :param str category: Category
        :param str metric: Metric
        :param str units: Units

        """
        if source:
            params['source'] = source

        if category:
            params['category'] = category

        if metric:
            params['metric'] = metric

        if units:
            params['units'] = units

        r = self.session.get('/statistics/%s' % (slug),
                             params=params, timeout=self.TIMEOUT)

        return r.json()

    def value(self, slug, source=None, category=None, metric=None, date=None, **params):
        """Get point values for a specific place

        :param str slug: Slug identifier (or shape id).
        :param list(str) source: Source(s)
        :param list(str) category: Category(s)
        :param list(str) metric: Metric(s)
        :param str date: Date
        """
        if source:

            if isinstance(source, string_types):
                source = [source]

            params['source'] = source

        if category:

            if isinstance(category, string_types):
                category = [category]

            params['category'] = category

        if metric:

            if isinstance(metric, string_types):
                metric = [metric]

            params['metric'] = metric

        if date:
            params['date'] = date

        r = self.session.get('/value/%s' % (slug),
                             params=params, timeout=self.TIMEOUT)

        return r.json()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings

from descarteslabs.client.auth import Auth
from descarteslabs.client.services.service import Service, ThirdPartyService


class Storage(Service):
    TIMEOUT = (9.5, 120)
    """Data Storage Service"""

    def __init__(self, url=None, auth=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if auth is None:
            auth = Auth()

        warnings.simplefilter('always', DeprecationWarning)
        if url is None:
            url = os.environ.get(
                "DESCARTESLABS_STORAGE_URL",
                "https://platform.descarteslabs.com/storage/v1"
            )

        self._gcs_upload_service = ThirdPartyService()

        super(Storage, self).__init__(url, auth=auth)

    def get_signed_url(self, key, storage_type='data'):
        r = self.session.get('/{storage_type}/get_signed_url/{key}'.format(storage_type=storage_type, key=key))
        r.raise_for_status()
        return r.content.decode('ascii')

    def get_upload_url(self, key, storage_type='data', **kwargs):
        r = self.session.get(
            '/{storage_type}/new_resumable_url/{key}'.format(storage_type=storage_type, key=key),
            params=kwargs
        )
        r.raise_for_status()
        return r.content.decode('ascii')

    def delete(self, key, storage_type='data'):
        """
        Delete the data stored at location `key` with storage type
        `storage_type`

        :param str key: A unique string mapped to an existing storage blob
        :param str storage_type: A type of data storage. Possible values: data", "tmp", "result".  Default: "data".

        """

        self.session.delete('/{storage_type}/{key}'.format(storage_type=storage_type, key=key))

    def set(self, key, value, storage_type='data'):
        """
        Store string `value` at location `key`, with storage type
        `storage_type`

        :param str key: A unique string mapped to an existing storage blob
        :param str value: bytes to be stored at location `key`
        :param str storage_type: A type of data storage. Possible values: "data", "tmp", "result".  Default: "data".
        """

        rurl = self.get_upload_url(key, storage_type=storage_type)

        self._gcs_upload_service.session.put(rurl, data=value)
        return

    def get(self, key, storage_type='data'):
        """
        Retrieve data stored at location `key`, with storage type
        `storage_type`

        :param str key: A unique string mapped to an existing storage blob
        :param str storage_type: A type of data storage. Possible values: "data", "tmp", "result".  Default: "data".

        Returns:
            The string that was stored at `key`

        """

        r = self.session.get('/{storage_type}/get/{key}'.format(storage_type=storage_type, key=key))
        r.raise_for_status()
        return r.content

    def list(self, prefix=None, storage_type='data'):
        """
        List keys that have been stored, with an optional `prefix` and `storage_type`.
        :param str prefix: A prefix match of keys returned.
        :param str storage_type: A type of data storage. Possible values: "data", "tmp", "result".  Default: "data".

        Returns:
            A list of keys.

        """

        r = self.session.get(
            '/{storage_type}/list'.format(storage_type=storage_type),
            params={'prefix': prefix}
        )
        r.raise_for_status()
        return r.json()

    def iter_list(self, prefix=None, storage_type='data'):
        """
        Yield keys that have been stored, with an optional `prefix` and `storage_type`.
        :param str prefix: A prefix match of keys returned.
        :param str storage_type: A type of data storage. Possible values: "data", "tmp", "result".  Default: "data".

        Returns:
            Yields keys in an iterable.

        """

        r = self.session.get(
            '/{storage_type}/list'.format(storage_type=storage_type),
            params={'prefix': prefix}
        )
        r.raise_for_status()
        for item in r.json():
            yield item
        while r.headers.get('X-NEXT', None) is not None:
            r = self.session.get(
                '/{storage_type}/list'.format(storage_type=storage_type),
                headers={'X-NEXT': r.headers['X-NEXT']},
                params={'prefix': prefix}
            )
            r.raise_for_status()
            for item in r.json():
                yield item

    def copy_from_bucket(self, src_bucket_name, src, dest, user_ns=None, storage_type='data'):
        """Copy a file from a google cloud storage bucket to your descartes
        data bucket. This requires that the dlstorage service account have
        access to your 3rd party bucket.
        :param str src_bucket_name: The name of the bucket from which you want to copy data.
        :param str src: The path to the file you want to copy inside the src_bucket_name` bucket.
        :param str dest: The new path of the file inside the descarteslabs data bucket.

        """
        if user_ns is None:
            user_ns = self.auth.namespace
        r = self.session.post(
            '/copy/{user_ns}/src/{src_bucket}/{src}/dest/{storage_type}/{dest}'.format(
                user_ns=user_ns,
                src_bucket=src_bucket_name,
                src=src,
                storage_type=storage_type,
                dest=dest
            ),
            json={}
        )
        return r.json()


storage = Storage()
storage_client = storage



#!/usr/bin/env python
# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .metadata import Metadata


def metadata_handler(args):
    metadata = Metadata()

    if args.url:
        metadata.url = args.url

    kwargs = {}

    if args.command == 'sources':
        sources = metadata.sources()

        print(json.dumps(sources, indent=2))

    if args.command == 'summary':
        if args.place:
            kwargs['place'] = args.place
        if args.start_datetime:
            kwargs['start_datetime'] = args.start_datetime
        if args.end_datetime:
            kwargs['end_datetime'] = args.end_datetime
        if args.geom:
            kwargs['geom'] = args.geom
        if args.bbox:
            kwargs['bbox'] = args.bbox

        summary = metadata.summary(**kwargs)

        print(json.dumps(summary))

    if args.command == 'search':
        if args.place:
            kwargs['place'] = args.place
        if args.start_datetime:
            kwargs['start_datetime'] = args.start_datetime
        if args.end_datetime:
            kwargs['end_datetime'] = args.end_datetime
        if args.geom:
            kwargs['geom'] = args.geom
        if args.limit:
            kwargs['limit'] = args.limit
        if args.offset:
            kwargs['offset'] = args.offset
        if args.bbox:
            kwargs['bbox'] = args.bbox

        search = metadata.search(**kwargs)

        print(json.dumps(search, indent=2))

    if args.command == 'ids':
        if args.place:
            kwargs['place'] = args.place
        if args.start_datetime:
            kwargs['start_datetime'] = args.start_datetime
        if args.end_datetime:
            kwargs['end_datetime'] = args.end_datetime
        if args.geom:
            kwargs['geom'] = args.geom
        if args.limit:
            kwargs['limit'] = args.limit
        if args.offset:
            kwargs['offset'] = args.offset
        if args.bbox:
            kwargs['bbox'] = args.bbox

        ids = metadata.ids(**kwargs)

        print(' '.join(ids))

    if args.command == 'get':
        get = metadata.get(args.argument)

        print(json.dumps(get))



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from warnings import warn, simplefilter
from six import string_types
from descarteslabs.client.services.service import Service
from descarteslabs.client.services.places import Places
from descarteslabs.client.auth import Auth
from descarteslabs.client.services.raster import Raster
from descarteslabs.client.services.metadata.metadata_filtering import \
    AndExpression, GLOBAL_PROPERTIES
from descarteslabs.common.dotdict import DotDict, DotList


OFFSET_DEPRECATION_MESSAGE = (
    "Keyword arg `offset` has been deprecated and will be removed in "
    "future versions of the library. "
)

SAT_ID_DEPRECATION_MESSAGE = (
    "Keyword arg `sat_id` has been deprecated and will be removed in "
    "future versions of the library. "
)

SOURCES_DEPRECATION_MESSAGE = (
    "Metadata.sources() has been deprecated and will be removed in "
    "future versions of the library. Please use "
    "Metadata.available_products() or Metadata.products() instead. "
)


class Metadata(Service):
    """
    Image Metadata Service

    Any methods that take start and end timestamps accept most common date/time
    formats as a string. If no explicit timezone is given, the timestamp is assumed
    to be in UTC. For example ``'2012-06-01'`` means June 1st 2012 00:00 in UTC,
    ``'2012-06-01 00:00+02:00'`` means June 1st 2012 00:00 in GMT+2.
    """

    TIMEOUT = (9.5, 120)

    properties = GLOBAL_PROPERTIES

    def __init__(self, url=None, auth=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if auth is None:
            auth = Auth()

        simplefilter('always', DeprecationWarning)
        if url is None:
            url = os.environ.get("DESCARTESLABS_METADATA_URL",
                                 "https://platform.descarteslabs.com/metadata/v1")

        super(Metadata, self).__init__(url, auth=auth)
        self._raster = Raster(auth=self.auth)

    def sources(self):
        warn(SOURCES_DEPRECATION_MESSAGE, DeprecationWarning)

        r = self.session.get('/sources')
        return DotList(r.json())

    def bands(
        self,
        products=None,
        limit=None,
        offset=None,
        wavelength=None,
        resolution=None,
        tags=None,
        bands=None,
        **kwargs
    ):
        """Search for imagery data bands that you have access to.

        :param list(str) products: A list of product(s) to return bands for.
        :param int limit: Number of results to return.
        :param int offset: Index to start at when returning results.
        :param float wavelength: A wavelength in nm e.g 700 that the band sensor must measure.
        :param int resolution: The resolution in meters per pixel e.g 30 of the data available in this band.
        :param list(str) tags: A list of tags that the band must have in its own tag list.


        """
        params = ['limit', 'offset', 'products',
                  'wavelength', 'resolution', 'tags']

        args = locals()
        kwargs = dict(kwargs, **{
            param: args[param]
            for param in params
            if args[param] is not None
        })

        r = self.session.post('/bands/search', json=kwargs)
        return DotList(r.json())

    def derived_bands(self, bands=None, require_bands=None, limit=None, offset=None, **kwargs):
        """Search for predefined derived bands that you have access to.

        :param list(str) bands: Limit the derived bands to ones that can be
                                computed using this list of spectral bands.
                                e.g ["red", "nir", "swir1"]
        :param bool require_bands: Control whether searched bands must contain
                                   all the spectral bands passed in the bands param.
                                   Defaults to False.
        :param int limit: Number of results to return.
        :param int offset: Index to start at when returning results.
        """
        params = ['bands', 'limit', 'offset']

        args = locals()
        kwargs = dict(kwargs, **{
            param: args[param]
            for param in params
            if args[param] is not None
        })

        r = self.session.post('/bands/derived/search', json=kwargs)
        return DotList(r.json())

    def get_bands_by_id(self, id_):
        """
        For a given source id, return the available bands.

        :param str id_: A :class:`Metadata` identifier.

        :return: A dictionary of band entries and their metadata.
        """
        r = self.session.get('/bands/id/{}'.format(id_))

        return DotDict(r.json())

    def products(self, bands=None, limit=None, offset=None, owner=None, text=None, **kwargs):
        """Search products that are available on the platform.

        :param list(str) bands: Band name(s) e.g ["red", "nir"] to filter products by.
                                Note that products must match all bands that are passed.
        :param int limit: Number of results to return.
        :param int offset: Index to start at when returning results.
        :param str owner: Filter products by the owner's uuid.
        :param str text: Filter products by string match.

        """
        params = ['limit', 'offset', 'bands', 'owner', 'text']

        args = locals()
        kwargs = dict(kwargs, **{
            param: args[param]
            for param in params
            if args[param] is not None
        })

        r = self.session.post('/products/search', json=kwargs)

        return DotList(r.json())

    def available_products(self):
        """Get the list of product identifiers you have access to.

        Example::
            >>> from descarteslabs.client.services import Metadata
            >>> products = Metadata().available_products()
            >>> products  # doctest: +SKIP
            ['landsat:LC08:PRE:TOAR']

        """
        r = self.session.get('/products')

        return DotList(r.json())

    def summary(self, products=None, sat_id=None, date='acquired', part=None,
                place=None, geom=None, start_datetime=None, end_datetime=None, cloud_fraction=None,
                cloud_fraction_0=None, fill_fraction=None, q=None, pixels=None,
                dltile=None, **kwargs):
        """Get a summary of the results for the specified spatio-temporal query.

        :param list(str) products: Product identifier(s).
        :param list(str) sat_id: Satellite identifier(s). *Deprecated*
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str part: Part of the date to aggregate over (e.g. `day`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_datetime: Desired starting timestamp, in any common format.
        :param str end_datetime: Desired ending timestamp, in any common format.
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param expr q: Expression for filtering the results. See :py:attr:`descarteslabs.utilities.properties`.
        :param bool pixels: Whether to include pixel counts in summary calculations.
        :param str dltile: A dltile key used to specify the resolution, bounds, and srs.

        Example usage::

            >>> from descarteslabs.client.services import Metadata
            >>> Metadata().summary(place='north-america_united-states_iowa', \
                    products=['landsat:LC08:PRE:TOAR'], start_datetime='2016-07-06', \
                    end_datetime='2016-07-07', part='hour', pixels=True)
            {
              'bytes': 93298309,
              'count': 1,
              'items': [
                {
                  'bytes': 93298309,
                  'count': 1,
                  'date': '2016-07-06T16:00:00',
                  'pixels': 250508160,
                  'timestamp': 1467820800
                }
              ],
              'pixels': 250508160,
              'products': ['landsat:LC08:PRE:TOAR']
            }
        """
        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            geom = json.dumps(shape['geometry'])

        if dltile is not None:
            if isinstance(dltile, string_types):
                dltile = self._raster.dltile(dltile)
            if isinstance(dltile, dict):
                geom = dltile['geometry']

        if isinstance(geom, dict):
            geom = json.dumps(geom)

        if sat_id:
            warn(SAT_ID_DEPRECATION_MESSAGE, DeprecationWarning)

            if isinstance(sat_id, string_types):
                sat_id = [sat_id]

            kwargs['sat_id'] = sat_id

        if products:
            if isinstance(products, string_types):
                products = [products]

            kwargs['products'] = products

        if date:
            kwargs['date'] = date

        if part:
            kwargs['part'] = part

        if geom:
            kwargs['geom'] = geom

        if start_datetime:
            kwargs['start_datetime'] = start_datetime

        if end_datetime:
            kwargs['end_datetime'] = end_datetime

        if cloud_fraction is not None:
            kwargs['cloud_fraction'] = cloud_fraction

        if cloud_fraction_0 is not None:
            kwargs['cloud_fraction_0'] = cloud_fraction_0

        if fill_fraction is not None:
            kwargs['fill_fraction'] = fill_fraction

        if q is not None:
            if not isinstance(q, list):
                q = [q]
            kwargs['query_expr'] = AndExpression(q).serialize()

        if pixels:
            kwargs['pixels'] = pixels

        r = self.session.post('/summary', json=kwargs)
        return DotDict(r.json())

    def search(self, products=None, sat_id=None, date='acquired', place=None,
               geom=None, start_datetime=None, end_datetime=None, cloud_fraction=None,
               cloud_fraction_0=None, fill_fraction=None, q=None, limit=100, offset=0,
               fields=None, dltile=None, sort_field=None, sort_order="asc", randomize=None,
               continuation_token=None, **kwargs):
        """Search metadata given a spatio-temporal query. All parameters are
        optional. For accessing more than 10000 results, see :py:func:`features`.

        :param list(str) products: Product Identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_datetime: Desired starting timestamp, in any common format.
        :param str end_datetime: Desired ending timestamp, in any common format.
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param expr q: Expression for filtering the results. See :py:attr:`descarteslabs.utilities.properties`.
        :param int limit: Number of items to return up to the maximum of 10000.
        :param int offset: Number of items to skip.
        :param list(str) fields: Properties to return.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param str sort_field: Property to sort on.
        :param str sort_order: Order of sort.
        :param bool randomize: Randomize the results. You may also use an `int` or `str` as an explicit seed.

        return: GeoJSON ``FeatureCollection``

        Example::

            >>> from descarteslabs.client.services import Metadata
            >>> scenes = Metadata().search(place='north-america_united-states_iowa', \
                                         products=['landsat:LC08:PRE:TOAR'], \
                                         start_datetime='2016-07-01', \
                                         end_datetime='2016-07-31T23:59:59')
            >>> len(scenes['features'])  # doctest: +SKIP
            2
        """
        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            geom = json.dumps(shape['geometry'])

        if dltile is not None:
            if isinstance(dltile, string_types):
                dltile = self._raster.dltile(dltile)
            if isinstance(dltile, dict):
                geom = dltile['geometry']

        if isinstance(geom, dict):
            geom = json.dumps(geom)

        kwargs.update({'date': date, 'limit': limit})

        if offset:
            warn(OFFSET_DEPRECATION_MESSAGE, DeprecationWarning)
            kwargs['offset'] = offset

        if sat_id:
            warn(SAT_ID_DEPRECATION_MESSAGE, DeprecationWarning)

            if isinstance(sat_id, string_types):
                sat_id = [sat_id]

            kwargs['sat_id'] = sat_id

        if products:
            if isinstance(products, string_types):
                products = [products]

            kwargs['products'] = products

        if geom:
            kwargs['geom'] = geom

        if start_datetime:
            kwargs['start_datetime'] = start_datetime

        if end_datetime:
            kwargs['end_datetime'] = end_datetime

        if cloud_fraction is not None:
            kwargs['cloud_fraction'] = cloud_fraction

        if cloud_fraction_0 is not None:
            kwargs['cloud_fraction_0'] = cloud_fraction_0

        if fill_fraction is not None:
            kwargs['fill_fraction'] = fill_fraction

        if fields is not None:
            kwargs['fields'] = fields

        if q is not None:
            if not isinstance(q, list):
                q = [q]
            kwargs['query_expr'] = AndExpression(q).serialize()

        if sort_field is not None:
            kwargs['sort_field'] = sort_field

        if sort_order is not None:
            kwargs['sort_order'] = sort_order

        if randomize is not None:
            kwargs['random_seed'] = randomize

        if continuation_token is not None:
            kwargs['continuation_token'] = continuation_token

        r = self.session.post('/search', json=kwargs)

        fc = {'type': 'FeatureCollection', "features": r.json()}

        if 'x-continuation-token' in r.headers:
            fc['properties'] = {
                'continuation_token': r.headers['x-continuation-token']}

        return DotDict(fc)

    def ids(self, products=None, sat_id=None, date='acquired', place=None,
            geom=None, start_datetime=None, end_datetime=None, cloud_fraction=None,
            cloud_fraction_0=None, fill_fraction=None, q=None, limit=100, offset=None,
            dltile=None, sort_field=None, sort_order=None, randomize=None, **kwargs):
        """Search metadata given a spatio-temporal query. All parameters are
        optional.

        :param list(str) products: Products identifier(s).
        :param list(str) sat_id: Satellite identifier(s).
        :param str date: The date field to use for search (e.g. `acquired`).
        :param str place: A slug identifier to be used as a region of interest.
        :param str geom: A GeoJSON or WKT region of interest.
        :param str start_datetime: Desired starting timestamp, in any common format.
        :param str end_datetime: Desired ending timestamp, in any common format.
        :param float cloud_fraction: Maximum cloud fraction, calculated by data provider.
        :param float cloud_fraction_0: Maximum cloud fraction, calculated by cloud mask pixels.
        :param float fill_fraction: Minimum scene fill fraction, calculated as valid/total pixels.
        :param expr q: Expression for filtering the results. See :py:attr:`descarteslabs.utilities.properties`.
        :param int limit: Number of items to return.
        :param int offset: Number of items to skip.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param str sort_field: Property to sort on.
        :param str sort_order: Order of sort.
        :param bool randomize: Randomize the results. You may also use an `int` or `str` as an explicit seed.

        :return: List of image identifiers.

        Example::

            >>> from descarteslabs.client.services import Metadata
            >>> ids = Metadata().ids(place='north-america_united-states_iowa', \
                                 products=['landsat:LC08:PRE:TOAR'], \
                                 start_datetime='2016-07-01', \
                                 end_datetime='2016-07-31T23:59:59')
            >>> len(ids)  # doctest: +SKIP
            1

            >>> ids  # doctest: +SKIP
            ['landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1', 'landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']

        """
        result = self.search(sat_id=sat_id, products=products, date=date,
                             place=place, geom=geom, start_datetime=start_datetime,
                             end_datetime=end_datetime, cloud_fraction=cloud_fraction,
                             cloud_fraction_0=cloud_fraction_0, fill_fraction=fill_fraction,
                             q=q, limit=limit, offset=offset, fields=[], dltile=dltile,
                             sort_field=sort_field, sort_order=sort_order, randomize=randomize, **kwargs)

        return DotList(feature['id'] for feature in result['features'])

    def features(self, products=None, sat_id=None, date='acquired', place=None,
                 geom=None, start_datetime=None, end_datetime=None, cloud_fraction=None,
                 cloud_fraction_0=None, fill_fraction=None, q=None, fields=None,
                 batch_size=1000, dltile=None, sort_field=None, sort_order='asc',
                 randomize=None, **kwargs):
        """Generator that efficiently scrolls through the search results.

        :param int batch_size: Number of features to fetch per request.

        :return: Generator of GeoJSON ``Feature`` objects.

        Example::

            >>> from descarteslabs.client.services import Metadata
            >>> features = Metadata().features("landsat:LC08:PRE:TOAR", \
                            start_datetime='2016-01-01', \
                            end_datetime="2016-03-01")
            >>> total = 0
            >>> for f in features: \
                    total += 1

            >>> total # doctest: +SKIP
            31898
        """

        continuation_token = None

        while True:
            result = self.search(sat_id=sat_id, products=products,
                                 date=date, place=place, geom=geom,
                                 start_datetime=start_datetime, end_datetime=end_datetime,
                                 cloud_fraction=cloud_fraction,
                                 cloud_fraction_0=cloud_fraction_0,
                                 fill_fraction=fill_fraction, q=q,
                                 fields=fields, limit=batch_size, dltile=dltile,
                                 sort_field=sort_field, sort_order=sort_order,
                                 randomize=randomize, continuation_token=continuation_token, **kwargs)

            if not result['features']:
                break

            for feature in result['features']:
                yield feature

            continuation_token = result['properties'].get('continuation_token')
            if not continuation_token:
                break

    def get(self, image_id):
        """Get metadata of a single image.

        :param str image_id: Image identifier.

        Example::

            >>> from descarteslabs.client.services import Metadata
            >>> meta = Metadata().get('landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1')
            >>> keys = list(meta.keys())
            >>> keys.sort()
            >>> keys
            ['acquired', 'area', 'bits_per_pixel', 'bright_fraction', 'bucket', 'cloud_fraction',
             'cloud_fraction_0', 'cs_code', 'descartes_version', 'file_md5s', 'file_sizes', 'files',
             'fill_fraction', 'geolocation_accuracy', 'geometry', 'geotrans', 'id', 'identifier', 'key',
             'processed', 'product', 'projcs', 'published', 'raster_size', 'reflectance_scale', 'roll_angle',
             'sat_id', 'solar_azimuth_angle', 'solar_elevation_angle', 'sw_version', 'terrain_correction',
             'tile_id']
        """
        r = self.session.get('/get/{}'.format(image_id))
        return DotDict(r.json())

    def get_by_ids(self, ids):
        """Get metadata for multiple images by id. The response contains found images in the
        order of the given ids. If no image exists for an id, that id is ignored.

        :param list(str) ids: Image identifiers.
        :return: List of image metadata.
        """
        r = self.session.post('/batch/images', json={'ids': ids})
        return DotList(r.json())

    def get_product(self, product_id):
        """Get information about a single product.

        :param str product_id: Product Identifier.

        """
        r = self.session.get('/products/{}'.format(product_id))
        return DotDict(r.json())

    def get_band(self, band_id):
        """Get information about a single product.

        :param str band_id: Band Identifier.

        """
        r = self.session.get('/bands/{}'.format(band_id))
        return DotDict(r.json())

    def get_derived_band(self, derived_band_id):
        """Get information about a single product.

        :param str derived_band_id: Derived band identifier.

        """
        r = self.session.get('/bands/derived/{}'.format(derived_band_id))
        return DotDict(r.json())



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

import mock
import responses

from descarteslabs.client.auth import Auth
from descarteslabs.client.services.tasks import CloudFunction, FutureTask, Tasks, as_completed

# flake8: noqa
public_token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJncm91cHMiOlsicHVibGljIl0sImlzcyI6Imh0dHBzOi8vZGVzY2FydGVzbGFicy5hdXRoMC5jb20vIiwic3ViIjoiZ29vZ2xlLW9hdXRoMnwxMTExMzg1NTY1MjQ4MTIzOTU3MTIiLCJhdWQiOiJaT0JBaTRVUk9sNWdLWklweHhsd09FZng4S3BxWGYyYyIsImV4cCI6OTk5OTk5OTk5OSwiaWF0IjoxNDc4MjAxNDE5fQ.sbSzD9ACNZvaxSgClZCnZMpee_p5MBaKV9uHZQonD6Q"


class ClientTestCase(unittest.TestCase):

    def setUp(self):
        url = "http://example.com"
        self.client = Tasks(url=url, auth=Auth(jwt_token=public_token, token_info_path=None))
        self.match_url = re.compile(url)

    def mock_response(self, method, json, status=200, **kwargs):
        responses.add(method, self.match_url, json=json, status=status, **kwargs)


class TasksTest(ClientTestCase):

    @responses.activate
    def test_new_group(self):
        def f():
            return True
        self.mock_response(responses.POST, {'id': 'foo'})
        group = self.client.new_group(f)
        self.assertEqual('foo', group.id)

    @responses.activate
    def test_iter_groups(self):
        self.mock_response(responses.GET, {'groups': [{'id': 'foo'}], 'continuation_token': 'continue'})
        self.mock_response(responses.GET, {'groups': [{'id': 'bar'}], 'continuation_token': None})
        groups = self.client.iter_groups()
        self.assertEqual(['foo', 'bar'], [group.id for group in groups])

    @responses.activate
    def test_new_task(self):
        self.mock_response(responses.POST, {'tasks': [{'id': 'foo'}]})
        tasks = self.client.new_task("group_id", arguments=('foo'), parameters={'bar': 'baz'})
        self.assertEqual('foo', tasks.tasks[0].id)

    @responses.activate
    def test_iter_task_results(self):
        self.mock_response(responses.GET, {'results': [{'id': 'foo'}], 'continuation_token': 'continue'})
        self.mock_response(responses.GET, {'results': [{'id': 'bar'}], 'continuation_token': None})
        results = self.client.iter_task_results("group_id")
        self.assertEqual(['foo', 'bar'], [result.id for result in results])

    @responses.activate
    @mock.patch.object(Tasks, "COMPLETION_POLL_INTERVAL_SECONDS", 0)
    def test_wait_for_completion(self):
        self.mock_response(responses.GET, {'id': 'foo', 'queue': {'pending': 3, 'successes': 0, 'failures': 0}})
        self.mock_response(responses.GET, {'id': 'foo', 'queue': {'pending': 0, 'successes': 2, 'failures': 1}})
        self.client.wait_for_completion('foo', show_progress=False)


class CloudFunctionTest(ClientTestCase):

    def setUp(self):
        super(CloudFunctionTest, self).setUp()
        self.function = CloudFunction("group_id", client=self.client)

    @responses.activate
    def test_call(self):
        self.mock_response(responses.POST, {'tasks': [{'id': 'foo'}]})
        task = self.function("foo", bar="baz")
        self.assertEqual(self.function.group_id, task.guid)
        self.assertEqual("foo", task.tuid)
        self.assertEqual(("foo",), task.args)
        self.assertEqual({"bar": "baz"}, task.kwargs)

    @responses.activate
    def test_map(self):
        self.mock_response(responses.POST, {'tasks': [{'id': 'foo'}, {'id': 'bar'}]})
        tasks = self.function.map(iter(["foo", "bar"]))
        self.assertEqual(["foo", "bar"], [task.tuid for task in tasks])
        self.assertEqual([("foo",), ("bar",)], [task.args for task in tasks])

    @responses.activate
    def test_map_multi(self):
        self.mock_response(responses.POST, {'tasks': [{'id': 'foo'}, {'id': 'bar'}]})
        tasks = self.function.map(iter(["foo", "bar"]), iter(["baz"]))
        self.assertEqual(["foo", "bar"], [task.tuid for task in tasks])
        self.assertEqual([("foo", "baz"), ("bar", None)], [task.args for task in tasks])


class FutureTaskTest(ClientTestCase):

    @responses.activate
    @mock.patch.object(Tasks, "COMPLETION_POLL_INTERVAL_SECONDS", 0)
    @mock.patch.object(Tasks, "TASK_RESULT_BATCH_SIZE", 3)
    def test_as_completed(self):
        tasks = [FutureTask("group_id", str(n), client=self.client) for n in range(5)]
        response1 = [{'id': str(n), 'result_type': 'json'} for n in range(3)]
        response2 = [{'id': str(n), 'result_type': 'json'} for n in range(3, 5)]
        self.mock_response(responses.POST, {'results': response1})
        self.mock_response(responses.POST, {'results': response2})
        completed_tasks = list(as_completed(tasks, show_progress=False))
        self.assertEqual(5, len(completed_tasks))
        self.assertEqual(list(range(5)), [int(r._task_result['id']) for r in completed_tasks])


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import shutil
import unittest
import json

import descarteslabs.client.addons as addons
from descarteslabs.client.addons import numpy as np
from descarteslabs.client.services.raster import Raster
from descarteslabs.client.services.places import Places


class TestRaster(unittest.TestCase):
    raster = None
    places = None

    @classmethod
    def setUpClass(cls):
        cls.raster = Raster()
        cls.places = Places()

    def test_raster(self):
        # TODO: Remove key test once keys are deprecated and removed.
        # test with scene key
        r = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertTrue("landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1_red-green-blue-alpha.tif" in r['files'])

        # test with scene id
        r = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertTrue("landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1_red-green-blue-alpha.tif" in r['files'])

    def test_raster_save(self):
        tmpdir = tempfile.mkdtemp()
        try:
            response = self.raster.raster(
                inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
                bands=['red', 'green', 'blue', 'alpha'],
                resolution=960, save=True,
                outfile_basename="{}/my-raster".format(tmpdir)
            )
            with open("{}/my-raster.tif".format(tmpdir), "rb") as f:
                f.seek(0, os.SEEK_END)
                length = f.tell()
            self.assertEqual(
                length,
                len(response['files']["{}/my-raster.tif".format(tmpdir)])
            )
        finally:
            shutil.rmtree(tmpdir)

    def test_ndarray(self):
        try:
            data, metadata = self.raster.ndarray(
                inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
                bands=['red', 'green', 'blue', 'alpha'],
                resolution=960,
            )
            self.assertEqual(data.shape, (249, 245, 4))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 4)
        except ImportError:
            pass

    def test_ndarray_single_band(self):
        try:
            data, metadata = self.raster.ndarray(
                inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
                bands=['red'],
                resolution=960,
            )
            self.assertEqual(data.shape, (249, 245))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_raster_blosc(self):
        r = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )

        old_blosc = addons.blosc
        addons.blosc = addons.ThirdParty("blosc")

        r2 = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
        )

        addons.blosc = old_blosc

        self.assertEqual(r, r2)

    def test_ndarray_blosc(self):
        r, meta = self.raster.ndarray(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
            align_pixels=True
        )

        old_blosc = addons.blosc
        addons.blosc = addons.ThirdParty("blosc")

        r2, meta2 = self.raster.ndarray(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            resolution=960,
            align_pixels=True
        )

        addons.blosc = old_blosc

        self.assertTrue((r == r2).all())
        self.assertEqual(meta, meta2)

    def test_stack_dltile(self):
        dltile = '128:16:960.0:15:-2:37'
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1',
                'landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1']

        stack, metadata = self.raster.stack(
            keys,
            dltile=dltile,
            bands=['red', 'green', 'blue', 'alpha'],
        )
        self.assertEqual(stack.shape, (2, 160, 160, 4))
        self.assertEqual(stack.dtype, np.uint16)
        self.assertEqual(len(metadata), 2)
        self.assertNotEqual(metadata[0], metadata[1])

    def test_stack_dltile_gdal_order(self):
        dltile = '128:16:960.0:15:-2:37'
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1',
                'landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1']

        stack, metadata = self.raster.stack(
            keys,
            dltile=dltile,
            bands=['red', 'green', 'blue', 'alpha'],
            order="gdal"
        )
        self.assertEqual(stack.shape, (2, 4, 160, 160))
        self.assertEqual(stack.dtype, np.uint16)
        self.assertEqual(len(metadata), 2)
        self.assertNotEqual(metadata[0], metadata[1])

    def test_stack_one_image(self):
        dltile = '128:16:960.0:15:-2:37'
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']

        stack, metadata = self.raster.stack(
            keys,
            dltile=dltile,
            bands=['red', 'green', 'blue', 'alpha'],
        )
        self.assertEqual(stack.shape, (1, 160, 160, 4))
        self.assertEqual(stack.dtype, np.uint16)
        self.assertEqual(len(metadata), 1)

    def test_stack_one_band(self):
        dltile = '128:16:960.0:15:-2:37'
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']

        stack, metadata = self.raster.stack(
            keys,
            dltile=dltile,
            bands=['red'],
        )
        self.assertEqual(stack.shape, (1, 160, 160, 1))
        self.assertEqual(stack.dtype, np.uint16)
        self.assertEqual(len(metadata), 1)

    def test_stack_one_band_gdal_order(self):
        dltile = '128:16:960.0:15:-2:37'
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']

        stack, metadata = self.raster.stack(
            keys,
            dltile=dltile,
            bands=['red'],
            order="gdal"
        )
        self.assertEqual(stack.shape, (1, 1, 160, 160))
        self.assertEqual(stack.dtype, np.uint16)

    def test_stack_res_cutline_utm(self):
        geom = {
            'coordinates': ((
                (-95.66055514862535, 41.24469400862013),
                (-94.74931826062456, 41.26199387228942),
                (-94.76311013534223, 41.95357639323731),
                (-95.69397431605952, 41.93542085595837),
                (-95.66055514862535, 41.24469400862013)
            ),),
            'type': 'Polygon'
        }
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1',
                'landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1']
        resolution = 960
        stack, metadata = self.raster.stack(
            keys,
            resolution=resolution,
            cutline=geom,
            srs="EPSG:32615",
            bounds=(277280.0, 4569600.0, 354080.0, 4646400.0),
            bands=['red', 'green', 'blue', 'alpha']
        )
        self.assertEqual(stack.shape, (2, 80, 80, 4))
        self.assertEqual(stack.dtype, np.uint16)

    def test_stack_res_cutline_wgs84(self):
        geom = {
            'coordinates': ((
                (-95.66055514862535, 41.24469400862013),
                (-94.74931826062456, 41.26199387228942),
                (-94.76311013534223, 41.95357639323731),
                (-95.69397431605952, 41.93542085595837),
                (-95.66055514862535, 41.24469400862013)
            ),),
            'type': 'Polygon'
        }
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1',
                'landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1']
        resolution = 960
        stack, metadata = self.raster.stack(
            keys,
            resolution=resolution,
            cutline=geom,
            srs="EPSG:32615",
            bounds=(-95.69397431605952, 41.24469400862013, -94.74931826062456, 41.95357639323731),
            bounds_srs="EPSG:4326",
            bands=['red', 'green', 'blue', 'alpha']
        )
        self.assertEqual(stack.shape, (2, 80, 84, 4))
        self.assertEqual(stack.dtype, np.uint16)

    def test_stack_underspecified(self):
        keys = ['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1']
        cutline = {
            'coordinates': ((
                (-95.66055514862535, 41.24469400862013),
                (-94.74931826062456, 41.26199387228942),
                (-94.76311013534223, 41.95357639323731),
                (-95.69397431605952, 41.93542085595837),
                (-95.66055514862535, 41.24469400862013)
            ),),
            'type': 'Polygon'
        }
        place = "north-america_united-states_iowa"
        bounds = (-95.69397431605952, 41.24469400862013, -94.74931826062456, 41.95357639323731)
        resolution = 960
        dimensions = (128, 128)
        srs = "EPSG:32615"

        with self.assertRaises(ValueError):
            self.raster.stack(keys)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, resolution=resolution)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, dimensions=dimensions)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, bounds=bounds)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, resolution=resolution, cutline=cutline)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, resolution=resolution, cutline=cutline, bounds=bounds)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, resolution=resolution, cutline=cutline, srs=srs)
        with self.assertRaises(ValueError):
            self.raster.stack(keys, resolution=resolution, place=place)

    def test_cutline_dict(self):
        shape = {
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[-95.2989209, 42.7999878], [-93.1167728, 42.3858464],
                                 [-93.7138666, 40.703737], [-95.8364984, 41.1150618],
                                 [-95.2989209, 42.7999878]]]
            }
        }
        try:
            data, metadata = self.raster.ndarray(
                inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
                bands=['red'],
                resolution=960,
                cutline=shape,
            )
            self.assertEqual(data.shape, (245, 238))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_cutline_str(self):
        shape = {
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[-95.2989209, 42.7999878], [-93.1167728, 42.3858464],
                                 [-93.7138666, 40.703737], [-95.8364984, 41.1150618],
                                 [-95.2989209, 42.7999878]]]
            }
        }
        try:
            data, metadata = self.raster.ndarray(
                inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
                bands=['red'],
                resolution=960,
                cutline=json.dumps(shape),
            )
            self.assertEqual(data.shape, (245, 238))
            self.assertEqual(data.dtype, np.uint16)
            self.assertEqual(len(metadata['bands']), 1)
        except ImportError:
            pass

    def test_thumbnail(self):
        r = self.raster.raster(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dimensions=[256, 256],
            scales=[[0, 4000]] * 4,
            output_format='PNG',
            data_type='Byte',
        )
        self.assertTrue("metadata" in r)
        self.assertTrue("files" in r)
        self.assertIsNotNone(r['files']['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1_red-green-blue-alpha.png'])

    def test_dltiles_from_place(self):
        iowa = self.places.shape('north-america_united-states_iowa', geom='low')
        iowa_geom = iowa['geometry']
        dltiles_feature_collection = self.raster.dltiles_from_shape(30.0, 2048, 16, iowa_geom)
        self.assertEqual(len(dltiles_feature_collection['features']), 58)

    def test_dltiles_from_latlon(self):
        dltile_feature = self.raster.dltile_from_latlon(45.0, -90.0, 30.0, 2048, 16)
        self.assertEqual(dltile_feature['properties']['key'], "2048:16:30.0:16:-4:81")

    def test_dltile(self):
        dltile_feature = self.raster.dltile("2048:16:30.0:16:-4:81")
        self.assertEqual(dltile_feature['properties']['key'], "2048:16:30.0:16:-4:81")

    def test_raster_dltile(self):
        dltile_feature = self.raster.dltile_from_latlon(41.0, -94.0, 30.0, 256, 16)
        arr, meta = self.raster.ndarray(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dltile=dltile_feature['properties']['key'],
        )
        self.assertEqual(arr.shape[0], 256 + 2 * 16)
        self.assertEqual(arr.shape[1], 256 + 2 * 16)
        self.assertEqual(arr.shape[2], 4)

    def test_raster_dltile_dict(self):
        dltile_feature = self.raster.dltile_from_latlon(41.0, -94.0, 30.0, 256, 16)
        arr, meta = self.raster.ndarray(
            inputs=['landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'],
            bands=['red', 'green', 'blue', 'alpha'],
            dltile=dltile_feature,
        )
        self.assertEqual(arr.shape[0], 256 + 2 * 16)
        self.assertEqual(arr.shape[1], 256 + 2 * 16)
        self.assertEqual(arr.shape[2], 4)


if __name__ == "__main__":
    unittest.main()



import cloudpickle
from hashlib import sha1
import logging
import time
import inspect


def cached(storage_client, minimum_runtime=0.0):
    def qualified_cached(f):
        FUNC_HASH = sha1("".join(inspect.getsourcelines(f)[0][1:]))\
            .hexdigest().encode("utf-8")

        def wrapper(*args, **kwargs):
            h = "/".join([
                FUNC_HASH,
                sha1(
                    cloudpickle.dumps((args, kwargs))
                ).hexdigest().encode("utf-8")
            ])

            # Check cache
            try:
                cached_result = storage_client.get(h, storage_type='cache')
                result = cloudpickle.loads(cached_result)
                logging.debug("Using cached result")
                return result
            except BaseException:
                pass

            t1 = -time.time()
            result = f(*args, **kwargs)
            t1 += time.time()

            if t1 > minimum_runtime:
                storage_client.set(h, cloudpickle.dumps(result), storage_type='cache')
                logging.debug("Cached result")

            return result
        return wrapper
    return qualified_cached



#!/usr/bin/env python
# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .places import Places


def places_handler(args):
    if args.geom == "None":
        args.geom = None

    places = Places()

    if args.url:
        places.url = args.url

    kwargs = {}

    if args.command == 'placetypes':
        placetypes = places.placetypes()

        print(json.dumps(placetypes, indent=2))

    if args.command == 'models':
        sources = places.sources()

        print(json.dumps(sources))

    if args.command == 'find':

        if args.placetype:
            kwargs['placetype'] = args.placetype

        find = places.find(args.argument, **kwargs)

        print(json.dumps(find, indent=2))

    if args.command == 'shape':
        kwargs['geom'] = args.geom
        shape = places.shape(args.argument, **kwargs)

        print(json.dumps(shape, indent=2))

    if args.command == 'prefix':
        if args.argument is None:
            raise RuntimeError("Please specify location in argument")

        if args.placetype is None:
            raise RuntimeError("Please specify placetype, e.g. -placetype region")

        kwargs['geom'] = args.geom
        kwargs['output'] = args.output
        kwargs['placetype'] = args.placetype

        prefix = places.prefix(args.argument, **kwargs)

        print(json.dumps(prefix, indent=2))



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ThirdParty(object):
    _package = None

    def __init__(self, package):
        self._package = package

    def __getattr__(self, name):
        raise ImportError("Please install the %s package" % self._package)

    def __dir__(self):
        raise ImportError("Please install the %s package" % self._package)

    def __call__(self, *args, **kwargs):
        raise ImportError("Please install the %s package" % self._package)


try:
    import numpy
except ImportError:
    numpy = ThirdParty("numpy")

try:
    import blosc
except ImportError:
    blosc = ThirdParty("blosc")

try:
    import shapely.geometry
except ImportError:
    shapely = ThirdParty("shapely")

try:
    import concurrent.futures
except ImportError:
    concurrent = ThirdParty("futures")

try:
    import matplotlib
except ImportError:
    matplotlib = ThirdParty("matplotlib")



import os
import six
import io

from tempfile import NamedTemporaryFile
from requests.exceptions import RequestException

from descarteslabs.client.addons import numpy as np
from descarteslabs.client.auth import Auth
from descarteslabs.client.exceptions import ServerError
from descarteslabs.client.services.metadata import Metadata
from descarteslabs.client.services.service import Service, ThirdPartyService


class Catalog(Service):
    """The Descartes Labs (DL) Catalog allows you to add georeferenced raster products
    into the Descartes Labs Platform. Catalog products can be used in other DL services
    like Raster and Tasks and Metadata.

    The entrypoint for using catalog is creating a Product using :meth:`Catalog.add_product`.
    After creating a product you should add band(s) to it using :meth:`Catalog.add_band`, and
    upload imagery using :meth:`Catalog.upload_image`.

    Bands define where and how data is encoded into imagery files, and how it should be displayed.
    Images define metadata about a specific groups of pixels (in the form of images), their georeferencing, geometry,
    coordinate system, and other pertinent information.
    """
    UPLOAD_NDARRAY_SUPPORTED_DTYPES = ['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'float32', 'float64']
    TIMEOUT = (9.5, 30)

    def __init__(self, url=None, auth=None, metadata=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if auth is None:
            auth = Auth()

        if metadata is None:
            self._metadata = Metadata(auth=auth)
        else:
            self._metadata = metadata

        self._gcs_upload_service = ThirdPartyService()

        if url is None:
            url = os.environ.get(
                "DESCARTESLABS_CATALOG_URL",
                "https://platform.descarteslabs.com/metadata/v1/catalog"
            )

        super(Catalog, self).__init__(url, auth=auth)

    def namespace_product(self, product_id):
        namespace = self.auth.namespace

        if not product_id.startswith(namespace):
            product_id = '{}:{}'.format(namespace, product_id)

        return product_id

    def own_products(self):
        return self._metadata.products(owner=self.auth.payload['sub'])

    def own_bands(self):
        return self._metadata.bands(owner=self.auth.payload['sub'])

    def _add_core_product(self, product_id, **kwargs):
        kwargs['id'] = product_id
        r = self.session.post('/core/products', json=kwargs)
        return r.json()

    def get_product(self, product_id, add_namespace=True):
        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.get('/products/{}'.format(product_id))
        return r.json()

    def add_product(self, product_id, title, description, add_namespace=True, **kwargs):
        """Add a product to your catalog.

        :param str product_id: (Required) A unique name for this product.
        :param str title: (Required) Official product title.
        :param str description: (Required) Information about the product,
                                why it exists, and what it provides.

        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.
        :param list(str) read: A list of groups, or user hashes to give read access to.
        :param int spectral_bands: Number of spectral bands the product has.
        :param list(str) native_bands: A list of the names of the native bands of this product
                                        (most applicable to satellite sensors).
        :param str start_datetime: ISO8601 compliant date string, indicating start of product data.
        :param str end_datetime: ISO8601 compliant date string, indicating end of product data.
        :param str notes: Any notes to relay about the product.
        :param str orbit: Type of orbit (satellite only).
        :param str processing_level: Way in which raw data has been processed if any.
        :param str resolution: Pixel density of the data, provide units.
        :param str revisit: How often an AOI can expect updated data.
        :param str sensor: Name of the sensor used.
        :param str swath: How large an area the sensor captures at a given time.

        :return: JSON API representation of the product.
        :rtype: dict
        """
        for k, v in locals().items():
            if k in ['title', 'description']:
                if v is None:
                    raise TypeError("required arg `{}` not provided".format(k))
                kwargs[k] = v

        kwargs['id'] = self.namespace_product(product_id) if add_namespace else product_id
        r = self.session.post('/products', json=kwargs)
        return r.json()

    def replace_product(
            self,
            product_id,
            title,
            description,
            add_namespace=True,
            set_global_permissions=False,
            **kwargs
    ):
        """Replace a product in your catalog with a new version.

        :param str product_id: (Required) A unique name for this product.
        :param str title: (Required) Official product title.
        :param str description: (Required) Information about the product,
                                why it exists, and what it provides.

        :param bool add_namespace: Add your user namespace to the product_id.
            This is only relevant to administrators modifying other users products. Defaults to True.
        :param list(str) read: A list of groups, or user hashes to give read access to.
        :param int spectral_bands: Number of spectral bands the product has.
        :param list(str) native_bands: A list of the names of the native bands of this product
                                        (most applicable to satellite sensors).
        :param str start_datetime: ISO8601 compliant date string, indicating start of product data.
        :param str end_datetime: ISO8601 compliant date string, indicating end of product data.
        :param str notes: Any notes to relay about the product.
        :param str orbit: Type of orbit (satellite only).
        :param str processing_level: Way in which raw data has been processed if any.
        :param str resolution: Pixel density of the data, provide units.
        :param str revisit: How often an AOI can expect updated data.
        :param str sensor: Name of the sensor used.
        :param str swath: How large an area the sensor captures at a given time.

        :return: JSON API representation of the product.
        :rtype: dict
        """
        for k, v in locals().items():
            if k in ['title', 'description']:
                if v is None:
                    raise TypeError("required arg `{}` not provided".format(k))
                kwargs[k] = v

        if add_namespace:
            product_id = self.namespace_product(product_id)

        params = None
        if set_global_permissions is True:
            params = {'set_global_permissions': 'true'}

        r = self.session.put('/products/{}'.format(product_id), json=kwargs, params=params)
        return r.json()

    def change_product(self, product_id, add_namespace=True, set_global_permissions=False, **kwargs):
        """Update a product to your catalog.

        :param str product_id: (Required) The ID of the product to change.
        :param list(str) read: A list of groups, or user hashes to give read access to.
        :param int spectral_bands: Number of spectral bands the product has.
        :param list(str) native_bands: A list of the names of the native bands of this product
                                        (most applicable to satellite sensors).

        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.
        :param bool set_global_permissions: Update permissions of all existing bands and products
                                            that belong to this product with the updated permission set
                                            specified in the `read` param. Default to false.

        :param str start_datetime: ISO8601 compliant date string, indicating start of product data.
        :param str end_datetime: ISO8601 compliant date string, indicating end of product data.
        :param str title: Official product title.
        :param str description: Information about the product,
                                why it exists, and what it provides.
        :param str notes: Any notes to relay about the product.
        :param str orbit: Type of orbit (satellite only).
        :param str processing_level: Way in which raw data has been processed if any.
        :param str resolution: Pixel density of the data, provide units.
        :param str revisit: How often an AOI can expect updated data.
        :param str sensor: Name of the sensor used.
        :param str swath: How large an area the sensor captures at a given time.

        :return: JSON API representation of the product.
        :rtype: dict
        """

        if add_namespace:
            product_id = self.namespace_product(product_id)

        params = None
        if set_global_permissions is True:
            params = {'set_global_permissions': 'true'}

        r = self.session.patch('/products/{}'.format(product_id), json=kwargs, params=params)
        return r.json()

    def remove_product(self, product_id, add_namespace=True, cascade=False):
        """Remove a product from the catalog.

        :param str product_id: ID of the product to remove.
        :param bool cascade: Force deletion of all the associated bands and images. BEWARE this cannot be undone.
        """
        if add_namespace:
            product_id = self.namespace_product(product_id)
        params = {'cascade': cascade}
        r = self.session.delete('/products/{}'.format(product_id), params=params)
        if r.headers['content-type'] == 'application/json':
            return r.json()

    def product_deletion_status(self, deletion_task_id):
        """Get the status of a long running product deletion job.

        :param str deletion_task_id: deletion_task ID returned from a call to :meth:`Catalog.remove_product`
                                    with a deletion_token.
        :return: document with information about product deletion progress.
        :rtype: dict
        """
        r = self.session.get('/products/deletion_tasks/{}'.format(deletion_task_id))
        return r.json()

    def get_band(self, product_id, name, add_namespace=True):
        """Get a band by name.

        :return: JSON API representation of the band.
        :rtype: dict
        """
        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.get('/products/{}/bands/{}'.format(product_id, name))
        return r.json()

    def _add_core_band(
            self,
            product_id,
            name,
            srcband=None,
            dtype=None,
            nbits=None,
            data_range=None,
            type=None,
            **kwargs
    ):
        for k, v in locals().items():
            if k in ['name', 'data_range', 'dtype', 'nbits', 'srcband', 'type']:
                if v is None:
                    raise TypeError("required arg `{}` not provided".format(k))
                kwargs[k] = v
        kwargs['id'] = name
        r = self.session.post('/core/products/{}/bands'.format(product_id), json=kwargs)
        return r.json()

    def add_band(
            self,
            product_id,
            name,
            add_namespace=True,
            srcband=None,
            dtype=None,
            nbits=None,
            data_range=None,
            type=None,
            **kwargs
    ):
        """Add a data band to an existing product.

        :param str product_id: (Required) Product to which this band will belong.
        :param str name: (Required) Name of this band.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.
        :param int jpx_layer: If data is processed to JPEG2000, which layer is the band in. Defaults to 0.
        :param int srcband: (Required) The 1-based index of the band in the jpx_layer specified.
        :param int srcfile: If the product was processed into multiple files,
                            which one is this in. Defaults to 0 (first file).
        :param str dtype: (Required) The data type used to store this band e.g Byte or Uint16 or Float32.
        :param int nbits: (Required) The number of bits of the dtype used to store this band.
        :param list(int) data_range: (Required) A list specifying the min and max values for the data in this band.
        :param str type: (Required) The data interpretation of the band. One of ['spectral', 'derived', 'mask', 'class']

        :param int nodata: Pixel value indicating no data available.
        :param list(str) read: A list of groups, or user hashes to give read access to.
                                     Defaults to the same as the parent product if not specified.
        :param str color: The color interpretation of this band.
                          One of ['Alpha', 'Black', 'Blue', 'Cyan', 'Gray', 'Green', 'Hue',
                          'Ligntness', 'Magenta', 'Palette', 'Red', 'Saturation',
                          'Undefined', 'YCbCr_Cb', 'YCbCr_Cr', 'YCbCr_Y', 'Yellow'].
                          Must be 'Alpha' if `type` is 'mask'.
                          Must be 'Palette' if `colormap_name` or `colormap` is set.
        :param str colormap_name: A named colormap to use for this band, one of
                                  ['plasma', 'magma', 'viridis', 'msw', 'inferno']
        :param list(list(str)) colormap: A custom colormap to use for this band. Takes a list of lists, where each
                                         nested list is a 4-tuple of rgba values to map pixels whose value is the index
                                         of the tuple. i.e the colormap [["100", "20", "200", "255"]] would map pixels
                                         whose value is 0 in the original band, to the rgba color vector at colormap[0].
                                         Less than 2^nbits 4-tuples may be provided, and omitted values
                                         will default map to black.

        :param str data_description: Description of band data.
        :param list(float) physical_range: If band represents a physical value e.g reflectance
                                           or radiance what are the possible range of those values.
        :param str data_unit: Units of the physical range e.g w/sr for radiance
        :param str data_unit_description: Description of the data unit.
        :param list(int) default_range: A good default scale for the band to use
                                        when rastering for display purposes.
        :param str description: Description of the band.
        :param str name_common: Standard name for band
        :param str name_vendor: What the vendor calls the band eg. B1
        :param int vendor_order: The index of the band in the vendors band tables.
                                 Useful for referencing the band to other processing
                                 properties like surface reflectance.
        :param str processing_level: Description of how the band was processed if at all.
        :param int res_factor: Scaling of this band relative to the native band resolution.
        :param int resolution: Resolution of this band.
        :param str resolution_unit: Unit of the resolution.
        :param float wavelength_center: Center position of wavelength.
        :param float wavelength_fwhm: Full width at half maximum value of the wavelength spread.
        :param float wavelength_min: Minimum wavelength this band is sensitive to.
        :param float wavelength_max: Maximum wavelength this band is sensitive to.
        :param str wavelength_unit: Units the wavelength is expressed in.

        :return: JSON API representation of the band.
        :rtype: dict
        """

        for k, v in locals().items():
            if k in ['name', 'data_range', 'dtype', 'nbits', 'srcband', 'type']:
                if v is None:
                    raise TypeError("required arg `{}` not provided".format(k))
                kwargs[k] = v
        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.post('/products/{}/bands'.format(product_id), json=kwargs)
        return r.json()

    def replace_band(
            self,
            product_id,
            name,
            srcband=None,
            dtype=None,
            nbits=None,
            data_range=None,
            add_namespace=True,
            type=None,
            **kwargs
    ):
        """Replaces an existing data band with a new document.

        :param str product_id: (Required) Product to which this band will belong.
        :param str name: (Required) Name of this band.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.
        :param int jpx_layer: If data is processed to JPEG2000, which layer is the band in. Defaults to 0.
        :param int srcband: (Required) The 1-based index of the band in the jpx_layer specified.
        :param int srcfile: If the product was processed into multiple files,
                            which one is this in. Defaults to 0 (first file).
        :param str dtype: (Required) The data type used to store this band e.g Byte or Uint16 or Float32.
        :param int nbits: (Required) The number of bits of the dtype used to store this band.
        :param list(int) data_range: (Required) A list specifying the min and max values for the data in this band.
        :param str type: (Required) The data interpretation of the band. One of ['spectral', 'derived', 'mask', 'class']

        :param int nodata: Pixel value indicating no data available.
        :param list(str) read: A list of groups, or user hashes to give read access to.
                                     Defaults to the same as the parent product if not specified.
        :param str color: The color interpretation of this band.
                          One of ['Alpha', 'Black', 'Blue', 'Cyan', 'Gray', 'Green', 'Hue',
                          'Ligntness', 'Magenta', 'Palette', 'Red', 'Saturation',
                          'Undefined', 'YCbCr_Cb', 'YCbCr_Cr', 'YCbCr_Y', 'Yellow'].
                          Must be 'Alpha' if `type` is 'mask'.
                          Must be 'Palette' if `colormap_name` or `colormap` is set.
        :param str colormap_name: A named colormap to use for this band, one of
                                  ['plasma', 'magma', 'viridis', 'msw', 'inferno']
        :param list(list(str)) colormap: A custom colormap to use for this band. Takes a list of lists, where each
                                         nested list is a 4-tuple of rgba values to map pixels whose value is the index
                                         of the tuple. i.e the colormap [["100", "20", "200", "255"]] would map pixels
                                         whose value is 0 in the original band, to the rgba color vector at colormap[0].
                                         Less than 2^nbits 4-tuples may be provided, and omitted values
                                         will default map to black.

        :param str data_description: Description of band data.
        :param list(float) physical_range: If band represents a physical value e.g reflectance
                                           or radiance what are the possible range of those values.
        :param str data_unit: Units of the physical range e.g w/sr for radiance
        :param str data_unit_description: Description of the data unit.
        :param list(int) default_range: A good default scale for the band to use
                                        when rastering for display purposes.
        :param str description: Description of the band.
        :param str name_common: Standard name for band
        :param str name_vendor: What the vendor calls the band eg. B1
        :param int vendor_order: The index of the band in the vendors band tables.
                                 Useful for referencing the band to other processing
                                 properties like surface reflectance.
        :param str processing_level: Description of how the band was processed if at all.
        :param int res_factor: Scaling of this band relative to the native band resolution.
        :param int resolution: Resolution of this band.
        :param str resolution_unit: Unit of the resolution.
        :param float wavelength_center: Center position of wavelength.
        :param float wavelength_fwhm: Full width at half maximum value of the wavelength spread.
        :param float wavelength_min: Minimum wavelength this band is sensitive to.
        :param float wavelength_max: Maximum wavelength this band is sensitive to.
        :param str wavelength_unit: Units the wavelength is expressed in.

        :return: JSON API representation of the band.
        :rtype: dict
        """

        for k, v in locals().items():
            if k in ['data_range', 'dtype', 'nbits', 'srcband', 'type']:
                if v is None:
                    raise TypeError("required arg `{}` not provided".format(k))
                kwargs[k] = v
        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.put('/products/{}/bands/{}'.format(product_id, name), json=kwargs)
        return r.json()

    def change_band(self, product_id, name, add_namespace=True, **kwargs):
        """Add a data band to an existing product.

        :param str product_id: (Required) Product to which this band belongs.
        :param str name: Name or id of band to modify.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.
        :param list(int) data_range: A list specifying the min and max values for the data in this band.
        :param str dtype: The data type used to store this band e.g Byte or Uint16 or Float32.
        :param int jpx_layer: If data is processed to JPEG2000, which layer is the band in. Use 0 for other formats.
        :param int nbits: The number of bits of the dtype used to store this band.
        :param int srcband: The 1-based index of the band in the jpx_layer specified.
        :param int srcfile: If the product was processed into multiple files,
                            which one is this in.
        :param str type: The data interpretation of the band. One of ['spectral', 'derived', 'mask', 'class']

        :param list(str) read: A list of groups, or user hashes to give read access to.
        :param str color: The color interpretation of this band.
                          One of ['Alpha', 'Black', 'Blue', 'Cyan', 'Gray', 'Green', 'Hue',
                          'Ligntness', 'Magenta', 'Palette', 'Red', 'Saturation',
                          'Undefined', 'YCbCr_Cb', 'YCbCr_Cr', 'YCbCr_Y', 'Yellow'].
                          Must be 'Alpha' if `type` is 'mask'.
                          Must be 'Palette' if `colormap_name` or `colormap` is set.
        :param str colormap_name: A named colormap to use for this band, one of
                                  ['plasma', 'magma', 'viridis', 'msw', 'inferno']
        :param list(list(str)) colormap: A custom colormap to use for this band. Takes a list of lists, where each
                                         nested list is a 4-tuple of rgba values to map pixels whose value is the index
                                         of the tuple. i.e the colormap [["100", "20", "200", "255"]] would map pixels
                                         whose value is 0 in the original band, to the rgba color vector at colormap[0].
                                         Less than 2^nbits 4-tuples may be provided, and omitted values
                                         will default map to black.

        :param str data_description: Description of band data.
        :param list(float) physical_range: If band represents a physical value e.g reflectance
                                           or radiance what are the possible range of those values.
        :param str data_unit: Units of the physical range e.g w/sr for radiance
        :param str data_unit_description: Description of the data unit.
        :param list(int) default_range: A good default scale for the band to use
                                        when rastering for display purposes.
        :param str description: Description of the band.
        :param str name_common: Standard name for band
        :param str name_vendor: What the vendor calls the band eg. B1
        :param int vendor_order: The index of the band in the vendors band tables.
                                 Useful for referencing the band to other processing
                                 properties like surface reflectance.
        :param int nodata:
        :param str processing_level: Description of how the band was processed if at all.
        :param int res_factor: Scaling of this band relative to the native band resolution.
        :param int resolution: Resolution of this band.
        :param str resolution_unit: Unit of the resolution.
        :param float wavelength_center: Center position of wavelength.
        :param float wavelength_fwhm: Full width at half maximum value of the wavelength spread.
        :param float wavelength_min: Minimum wavelength this band is sensitive to.
        :param float wavelength_max: Maximum wavelength this band is sensitive to.
        :param str wavelength_unit: Units the wavelength is expressed in.

        :return: JSON API representation of the band.
        :rtype: dict
        """
        if add_namespace:
            product_id = self.namespace_product(product_id)
        r = self.session.patch('/products/{}/bands/{}'.format(product_id, name), json=kwargs)
        return r.json()

    def remove_band(self, product_id, name, add_namespace=True):
        if add_namespace:
            product_id = self.namespace_product(product_id)
        self.session.delete('/products/{}/bands/{}'.format(product_id, name))

    def get_image(self, product_id, image_id, add_namespace=True):
        """Get a single image metadata entry.

        :return: JSON API representation of the image.
        :rtype: dict
        """

        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.get('/products/{}/images/{}'.format(product_id, image_id))
        return r.json()

    def add_image(self, product_id, image_id, add_namespace=True, **kwargs):
        """Add an image metadata entry to a product.

        :param str product_id: (Required) Product to which this image belongs.
        :param str image_id: (Required) New image's id = <product_id>:<image_id>.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.

        :param list(str) read: A list of groups, or user hashes to give read access to.
                                     Defaults to the same as the parent product if not specified.
        :param int absolute_orbit: Orbit number since mission start.
        :param str acquired: Date the imagery was acquired
        :param str archived: Date the imagery was archived.
        :param float area: Surface area the image covers
        :param float azimuth_angle:
        :param float azimuth_angle_1:
        :param list(float) bits_per_pixel: Average bits of data per pixel per band.
        :param float bright_fraction: Fraction of the image that has reflectance greater than .4 in the blue band.
        :param str bucket: Name of Google Cloud Bucket. Must be public bucket or Descartes Labs user bucket.
        :param str catalog_id:
        :param float cirrus_fraction: Fraction of pixel which are distorted by cirrus clouds.
        :param float cloud_fraction: Fraction of pixels which are obscured by clouds.
        :param float cloud_fraction_0: Fraction of pixels which are obscured by clouds.
        :param str cs_code: Spatial coordinate system code eg. `EPSG:4326`
        :param datastrip_id: ID of the data strip this image was taken in.
        :param float degraded_fraction_0:
        :param str descartes_version:
        :param str directory: Subdirectory location.
        :param float duration: duration of the scan in seconds
        :param int duration_days:
        :param float earth_sun_distance: Earth sun distance at time of image capture.
        :param list(str) files: Names of the files this image is stored in.
        :param list(str) file_md5s: File integrity checksums.
        :param list(int) file_sizes: Number of bytes of each file
        :param float fill_fraction: Fraction of this image which has data.
        :param float geolocation_accuracy:
        :param str|dict geometry: GeoJSON representation of the image coverage.
        :param float geotrans: Geographic Translator values.
        :param float gpt_time:
        :param str identifier: Vendor scene identifier.
        :param int ifg_tdelta_days:
        :param float incidence_angle: Sensor incidence angle.
        :param str pass: On which pass was this image taken.
        :param str processed: Date which this image was processed.
        :param str proj4: proj.4 transformation parameters
        :param str projcs: Projection coordinate system.
        :param str published: Date the image was published.
        :param str radiometric_version:
        :param list(int) raster_size: Dimensions of image in pixels in (width, height).
        :param list(float) reflectance_scale: Scale factors converting TOA radiances to TOA reflectances
        :param list(float) reflectance_scale_1:
        :param int relative_orbit: Orbit number in revisit cycle.
        :param float roll_angle:
        :param str safe_id: Standard Archive Format for Europe.
        :param str sat_id: Satellite identifier.
        :param float scan_gap_interpolation:
        :param float solar_azimuth_angle:
        :param float solar_azimuth_angle_0:
        :param float solar_azimuth_angle_1:
        :param float solar_elevation_angle:
        :param float solar_elevation_angle_0:
        :param float solar_elevation_angle_1:
        :param str tile_id:
        :param float view_angle:
        :param float view_angle_1:
        :param dict extra_properties: User defined custom properties for this image.
                Only 10 keys are allowed. The dict can only map strings to primitive types (str -> str|float|int).

        :return: JSON API representation of the image.
        :rtype: dict
        """

        if add_namespace:
            product_id = self.namespace_product(product_id)
        kwargs['id'] = image_id
        r = self.session.post('/products/{}/images'.format(product_id), json=kwargs)
        return r.json()

    def replace_image(self, product_id, image_id, add_namespace=True, **kwargs):
        """Replace image metadata with a new version.

        :param str product_id: (Required) Product to which this image belongs.
        :param str image_id: (Required) ID of the image to replace.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.

        :param list(str) read: A list of groups, or user hashes to give read access to.
                                     Defaults to the same as the parent product if not specified.
        :param int absolute_orbit: Orbit number since mission start.
        :param str acquired: Date the imagery was acquired
        :param str archived: Date the imagery was archived.
        :param float area: Surface area the image covers
        :param float azimuth_angle:
        :param float azimuth_angle_1:
        :param list(float) bits_per_pixel: Average bits of data per pixel.
        :param float bright_fraction: Fraction of the image that has reflectance greater than .4 in the blue band.
        :param str bucket: Name of Google Cloud Bucket. Must be public bucket or Descartes Labs user bucket.
        :param str catalog_id:
        :param float cirrus_fraction: Fraction of pixel which are distorted by cirrus clouds.
        :param float cloud_fraction: Fraction of pixels which are obscured by clouds.
        :param float cloud_fraction_0: Fraction of pixels which are obscured by clouds.
        :param str cs_code: Spatial coordinate system code eg. `EPSG:4326`
        :param datastrip_id: ID of the data strip this image was taken in.
        :param float degraded_fraction_0:
        :param str descartes_version:
        :param str directory: Subdirectory location.
        :param float duration:
        :param int duration_days:
        :param float earth_sun_distance: Earth sun distance at time of image capture.
        :param list(str) files: Names of the files this image is stored in.
        :param list(str) file_md5s: File integrity checksums.
        :param list(int) file_sizes: Number of bytes of each file
        :param float fill_fraction: Fraction of this image which has data.
        :param float geolocation_accuracy:
        :param str|dict geometry: GeoJSON representation of the image coverage.
        :param float geotrans: Geographic Translator values.
        :param float gpt_time:
        :param str identifier: Vendor scene identifier.
        :param int ifg_tdelta_days:
        :param float incidence_angle: Sensor incidence angle.
        :param str pass: On which pass was this image taken.
        :param str processed: Date which this image was processed.
        :param str proj4: proj.4 transformation parameters
        :param str projcs: Projection coordinate system.
        :param str published: Date the image was published.
        :param str radiometric_version:
        :param list(int) raster_size: Dimensions of image in pixels in (width, height).
        :param list(float) reflectance_scale: Scale factors converting TOA radiances to TOA reflectances
        :param list(float) reflectance_scale_1:
        :param int relative_orbit: Orbit number in revisit cycle.
        :param float roll_angle:
        :param str safe_id: Standard Archive Format for Europe.
        :param str sat_id: Satellite identifier.
        :param float scan_gap_interpolation:
        :param float solar_azimuth_angle:
        :param float solar_azimuth_angle_0:
        :param float solar_azimuth_angle_1:
        :param float solar_elevation_angle:
        :param float solar_elevation_angle_0:
        :param float solar_elevation_angle_1:
        :param str tile_id:
        :param float view_angle:
        :param float view_angle_1:

        :return: JSON API representation of the image.
        :rtype: dict
        """

        if add_namespace:
            product_id = self.namespace_product(product_id)

        r = self.session.put('/products/{}/images/{}'.format(product_id, image_id), json=kwargs)
        return r.json()

    def _add_core_image(self, product_id, image_id, **kwargs):
        kwargs['id'] = image_id
        r = self.session.post('/core/products/{}/images'.format(product_id), json=kwargs)
        return r.json()

    def change_image(self, product_id, image_id, add_namespace=True, **kwargs):
        """Add an image metadata entry to a product.

        :param str product_id: (Required) Product to which this image belongs.
        :param str image_id: (Required) ID of the image to modify.
        :param bool add_namespace: Add your user namespace to the product_id.
                                   This is only relevant to administrators modifying
                                   other users products. Defaults to True.

        :param list(str) read: A list of groups, or user hashes to give read access to.
        :param int absolute_orbit: Orbit number since mission start.
        :param str acquired: Date the imagery was acquired
        :param str archived: Date the imagery was archived.
        :param float area: Surface area the image covers
        :param float azimuth_angle:
        :param float azimuth_angle_1:
        :param list(float) bits_per_pixel: Average bits of data per pixel.
        :param float bright_fraction: Fraction of the image that has reflectance greater than .4 in the blue band.
        :param str bucket: Name of Google Cloud Bucket. Must be public bucket or Descartes Labs user bucket.
        :param str catalog_id:
        :param float cirrus_fraction: Fraction of pixel which are distorted by cirrus clouds.
        :param float cloud_fraction: Fraction of pixels which are obscured by clouds.
        :param float cloud_fraction_0: Fraction of pixels which are obscured by clouds.
        :param str cs_code: Spatial coordinate system code eg. `EPSG:4326`
        :param datastrip_id: ID of the data strip this image was taken in.
        :param float degraded_fraction_0:
        :param str descartes_version:
        :param str directory: Subdirectory location.
        :param float duration:
        :param int duration_days:
        :param float earth_sun_distance: Earth sun distance at time of image capture.
        :param list(str) files: Names of the files this image is stored in.
        :param list(str) file_md5s: File integrity checksums.
        :param list(int) file_sizes: Number of bytes of each file
        :param float fill_fraction: Fraction of this image which has data.
        :param float geolocation_accuracy:
        :param str|dict geometry: GeoJSON representation of the image coverage.
        :param float geotrans: Geographic Translator values.
        :param float gpt_time:
        :param str identifier: Vendor scene identifier.
        :param int ifg_tdelta_days:
        :param float incidence_angle: Sensor incidence angle.
        :param str pass: On which pass was this image taken.
        :param str processed: Date which this image was processed.
        :param str proj4: proj.4 transformation parameters
        :param str projcs: Projection coordinate system.
        :param str published: Date the image was published.
        :param str radiometric_version:
        :param list(int) raster_size: Dimensions of image in pixels (width, height).
        :param list(float) reflectance_scale: Scale factors converting TOA radiances to TOA reflectances
        :param list(float) reflectance_scale_1:
        :param int relative_orbit: Orbit number in revisit cycle.
        :param float roll_angle:
        :param str safe_id: Standard Archive Format for Europe.
        :param str sat_id: Satellite identifier.
        :param float scan_gap_interpolation:
        :param float solar_azimuth_angle:
        :param float solar_azimuth_angle_0:
        :param float solar_azimuth_angle_1:
        :param float solar_elevation_angle:
        :param float solar_elevation_angle_0:
        :param float solar_elevation_angle_1:
        :param str tile_id:
        :param float view_angle:
        :param float view_angle_1:

        :return: JSON API representation of the image.
        :rtype: dict
        """

        if add_namespace:
            product_id = self.namespace_product(product_id)
        r = self.session.patch('/products/{}/images/{}'.format(product_id, image_id), json=kwargs)
        return r.json()

    def remove_image(self, product_id, image_id, add_namespace=True):
        if add_namespace:
            product_id = self.namespace_product(product_id)
        self.session.delete('/products/{}/images/{}'.format(product_id, image_id))

    def upload_image(self, files, product_id, metadata=None, multi=False, image_id=None, **kwargs):
        """Upload an image for a product you own.

        :param str|file|list(str)|list(file) files: (Required) a reference to the file to upload.
        :param str product_id: (Required) The id of the product this image belongs to.
        :param dict metadata: Image metadata to use instead of the computed default values.
            see :meth:`Catalog.add_image` for allowed keys.
        """

        if metadata is None:
            metadata = {}
        metadata.update(kwargs)
        if multi is True:
            if not hasattr(files, '__iter__'):
                raise ValueError("Using `multi=True` requires `files` to be iterable")
            elif image_id is None:
                raise ValueError("Using `multi=True` requires `image_id` to be specified")
            else:
                upload = self._do_multi_file_upload(files, product_id, image_id, metadata)
        else:
            upload = self._do_upload(files, product_id, metadata=metadata)
        if upload[0]:
            raise upload[2]

    def upload_ndarray(
            self,
            ndarray,
            product_id,
            image_id,
            proj4=None,
            wkt_srs=None,
            geotrans=None,
            raster_meta=None,
            **kwargs
    ):
        """Upload an ndarray with georeferencing information.

        :param ndarray ndarray: (Required) A numpy ndarray with image data. If you are providing a multi-band image
            it should have 3 dimensions and the 3rd dimension of the array should index the bands. The dtype of the
            ndarray must also be one of the following:
            ['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'float32', 'float64']
        :param str product_id: (Required) The id of the product this image belongs to.
        :param str image_id: (Required) Resulting image's id = <product_id>:<image_id>.
        :param str proj4: (One of proj4 or wkt_srs is required) A proj4 formatted string representing the
            spatial reference system used by the image.
        :param str wkt_srs: (One of proj4 or wkt_srs is required) A well known text string representing the
            spatial reference system used by the image.
        :param list(float) geotrans: (Required) The 6 number geographic transform of the image. Maps pixel coordinates
            to coordinates in the specified spatial reference system.
        :param dict raster_meta: Metadata returned from the :meth:`descarteslabs.client.services.raster.Raster.ndarray`
            request which generated the initial data for the :param ndarray: being uploaded. Passing :param geotrans:
            and :param wkt_srs: is unnecessary in this case.

        .. note:: Only one of `proj4` or `wkt_srs` can be provided.
        """
        if ndarray.dtype.name not in self.UPLOAD_NDARRAY_SUPPORTED_DTYPES:
            raise TypeError("{} is not in supported dtypes {}".format(ndarray.dtype.name,
                                                                      self.UPLOAD_NDARRAY_SUPPORTED_DTYPES))

        metadata = kwargs
        metadata.setdefault('process_controls', {}).update({'upload_type': 'ndarray'})
        if raster_meta is not None:
            geotrans = raster_meta.get('geoTransform')
            wkt_srs = raster_meta.get('coordinateSystem', {}).get('wkt')
        for arg in ['image_id', 'proj4', 'wkt_srs', 'geotrans']:
            if locals()[arg] is not None:
                kwargs[arg] = locals()[arg]
        with NamedTemporaryFile() as tmp:
            np.save(tmp, ndarray, allow_pickle=False)
            # From tempfile docs:
            # Whether the name can be used to open the file a second time, while
            # the named temporary file is still open, varies across platforms
            # (it can be so used on Unix; it cannot on Windows NT or later)
            #
            # We close the underlying file object so _do_upload can open the path again
            # in a cross platform compatible way.
            # When leaving the context manager the tempfile wrapper will still cleanup
            # and unlink the file descriptor.
            tmp.file.close()
            upload = self._do_upload(tmp.name, product_id, metadata=metadata)
            if upload[0]:
                raise upload[2]

    def upload_results(
            self,
            product_id,
            limit=100,
            offset=None,
            status=None,
            updated=None,
            created=None,
            continuation_token=None,
    ):
        """Get result information for debugging your uploads.

        :param str product_id: Product ID to get upload results for.
        :param int limit: Number of results to get, useful for paging.
        :param int offset: Start of results to get, useful for paging.
        :param str status: Filter results by status, values are ["SUCCESS", "FAILURE"]
        :param str|int updated: Unix timestamp or ISO8601 formatted date for filtering results updated after this time.
        :param str|int created: Unix timestamp or ISO8601 formatted date for filtering results created after this time.

        :return: A list of upload result objects.
        :rtype: list
        """
        kwargs = {'limit': limit}
        for arg in ['offset', 'status', 'updated', 'created', 'continuation_token']:
            if locals()[arg] is not None:
                kwargs[arg] = locals()[arg]
        results = self.session.post(
            '/products/{}/uploads'.format(product_id),
            json=kwargs
        )
        return results.json()

    def iter_upload_results(
            self,
            product_id,
            status=None,
            updated=None,
            created=None,
    ):
        """Get result information for debugging your uploads.

        :param str product_id: Product ID to get upload results for.
        :param str status: Filter results by status, values are ["SUCCESS", "FAILURE"]
        :param str|int updated: Unix timestamp or ISO8601 formatted date for filtering results updated after this time.
        :param str|int created: Unix timestamp or ISO8601 formatted date for filtering results created after this time.

        :return: iterator to upload results.
        :rtype: generator
        """
        continuation_token = None
        kwargs = {}
        for arg in ['status', 'updated', 'created']:
            if locals()[arg] is not None:
                kwargs[arg] = locals()[arg]
        while True:
            page = self.upload_results(product_id, continuation_token=continuation_token, **kwargs)
            for res in page['data']:
                yield res
            continuation_token = page['meta']['continuation_token']
            if continuation_token is None:
                break

    def upload_result(self, product_id, upload_id):
        """Get one upload result with the processing logs.

        This is useful for debugging failed uploads.
        :param str product_id: Product ID to get upload result for.
        :param str upload_id: ID of specific upload to get a result record of, includes the run logs.

        :return: One upload result with run logs.
        :rtype: dict
        """
        result = self.session.get('/products/{}/uploads/{}'.format(product_id, upload_id))
        return result.json()

    def _do_multi_file_upload(self, files, product_id, image_id, metadata):
        file_keys = [os.path.basename(_f) for _f in files]
        process_controls = metadata.setdefault('process_controls', {'upload_type': 'file'})
        multi_file_args = {
            'multi_file': {
                'image_files': file_keys,
                'image_id': image_id,
            }
        }
        process_controls.update(multi_file_args)
        for _file in files:
            upload = self._do_upload(_file, product_id, metadata=metadata)
            if upload[0]:
                return upload
        else:
            return upload

    def _do_upload(self, file_ish, product_id, metadata=None):
        # kwargs are treated as metadata fields and restricted to primitives
        # for the key val pairs.
        fd = None
        product_id = self.namespace_product(product_id)

        if metadata is None:
            metadata = {}
        metadata.setdefault('process_controls', {'upload_type': 'file'})

        if isinstance(file_ish, io.IOBase):
            if 'b' not in file_ish.mode:
                file_ish = io.open(file_ish.name, 'rb')
            fd = file_ish
        elif isinstance(file_ish, six.string_types) and os.path.exists(file_ish):
            fd = io.open(file_ish, 'rb')
        else:
            return 1, file_ish, Exception(
                'Could not handle file: `{}` pass a valid path or open IOBase instance'.format(file_ish)
            )
        try:
            r = self.session.post(
                '/products/{}/images/upload/{}'.format(
                    product_id,
                    metadata.pop('image_id', None) or os.path.basename(fd.name)
                ),
                json=metadata
            )
            upload_url = r.text
            r = self._gcs_upload_service.session.put(upload_url, data=fd)
        except (ServerError, RequestException) as e:
            return 1, fd.name, e
        finally:
            fd.close()

        return 0, fd.name, ''


catalog = Catalog()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
from descarteslabs.client.services.metadata import Metadata
from descarteslabs.client.services.places import Places
from descarteslabs.client.services.raster import Raster
from descarteslabs.client.services.storage import Storage, cached
from descarteslabs.client.services.catalog import Catalog
from descarteslabs.client.services.tasks import AsyncTasks, Tasks, FutureTask, CloudFunction

__all__ = ["Raster", "Metadata", "Places", "Storage", "Catalog", "AsyncTasks", "Tasks", "FutureTask", "CloudFunction"]



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import stat
import json
import six
from six.moves import input

from descarteslabs.client.auth.auth import Auth, base64url_decode, makedirs_if_not_exists, DEFAULT_TOKEN_INFO_PATH
from descarteslabs.client.version import __version__


def auth_handler(args):
    auth = Auth()

    if args.command == 'login':

        print(
            'Follow this link to login https://iam.descarteslabs.com/auth/login?refresh_token=true&destination=/auth/refresh_token')   # NOQA

        s = input('...then come back here and paste the generated token: ')
        if isinstance(s, six.text_type):
            s = s.encode('utf-8')

        if s:

            token_info = json.loads(base64url_decode(s).decode('utf-8'))
            if "refresh_token" not in token_info:  # TODO(justin) legacy for previous IDP
                token_info['refresh_token'] = token_info.get("client_secret")

            token_info_directory = os.path.dirname(DEFAULT_TOKEN_INFO_PATH)
            makedirs_if_not_exists(token_info_directory)

            os.chmod(token_info_directory, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

            with open(DEFAULT_TOKEN_INFO_PATH, 'w+') as fp:
                json.dump(token_info, fp)

            os.chmod(DEFAULT_TOKEN_INFO_PATH, stat.S_IRUSR | stat.S_IWUSR)

            # Get a fresh Auth token
            auth = Auth()

            name = auth.payload['name']

            print('Welcome, %s!' % name)

    if args.command == 'token':
        print(auth.token)

    if args.command == 'name':
        auth.token
        print(auth.payload['name'])

    if args.command == 'groups':
        auth.token
        print(json.dumps(auth.payload['groups']))

    if args.command == 'payload':
        auth.token
        print(auth.payload)

    if args.command == 'env':
        auth.token
        print('%s=%s' % ('CLIENT_ID', auth.client_id))
        print('%s=%s' % ('CLIENT_SECRET', auth.client_secret))
        print('%s=%s' % ('REFRESH_TOKEN', auth.refresh_token))

    if args.command == 'version':
        print(__version__)
